'''Particular color values which occur often enough or are unique enough to have distinct names'''

__author__ = 'Timotej Bernat'
__email__ = 'timotej.bernat@colorado.edu'

BLACK = (0, 0, 0)
WHITE = (255, 255, 255)

RED   = (255, 0, 0)
GREEN = (0, 255, 0)
BLUE  = (0, 0, 255)