'''Tools for working with ordered sequences (i.e. collections of elements with definite length and order)'''

__author__ = 'Timotej Bernat'
__email__ = 'timotej.bernat@colorado.edu'

from .seqops import * # backwards-compatible change which also makes some utils more easily accessible