'''General-purpose utilities related to SMILES and SMARTS string manipulations'''

__author__ = 'Timotej Bernat'
__email__ = 'timotej.bernat@colorado.edu'

from .primitives import is_valid_SMILES, is_valid_SMARTS, Smiles, Smarts